﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=8.8
@EndOfDesignText@
'This is a conversion of Nikolau Chernov's c++ code to B4X code. c++ Source: https://people.cas.uab.edu/~mosya/cl/CircleFitByTaubin.cpp

'      Circle fit To a given set of data points (in 2D)
'      
'      This Is an algebraic fit, due To Taubin, based on the journal article
'     
'      G. Taubin, "Estimation Of Planar Curves, Surfaces And Nonplanar
'                  Space Curves Defined By Implicit Equations, With 
'                  Applications To Edge And Range Image Segmentation",
'                  IEEE Trans. PAMI, Vol. 13, pages 1115-1138, (1991)
'
'      Input:  data     - the class of data (contains the given points):
'		
'	      DataN   - the number of data points
'	      data.X[] - the Array of X-coordinates
'	      data.Y[] - the Array of Y-coordinates
'     
'     Output:	       
'               circle - parameters of the fitting circle:
'		        
'	       circle.a - the X-coordinate of the center of the fitting circle
'	       circle.b - the Y-coordinate of the center of the fitting circle
' 	       circle.r - the radius of the fitting circle
' 	       circle.s - the root mean square error (the estimate of sigma)
' 	       circle.j - the total number of iterations
'             
'     The method Is based on the minimization of the function
'     
'                  sum [(x-a)^2 + (y-b)^2 - R^2]^2
'              F = -------------------------------
'                      sum [(x-a)^2 + (y-b)^2]
'                 
'     This method Is more balanced than the simple Kasa fit.
'        
'     It works well whether data points are sampled along an entire circle Or
'     along a small arc. 
'     
'     It still has a small bias And its statistical accuracy Is slightly
'     lower than that of the geometric fit (minimizing geometric distances),
'     but slightly higher than that of the very similar Pratt fit. 
'     Besides, the Taubin fit Is slightly simpler than the Pratt fit
'     
'     It provides a very good initial guess For a subsequent geometric fit. 
'     
'       Nikolai Chernov  (September 2012)

Sub Class_Globals
	Private dataMeanX, dataMeanY As Double
	Private dataN As Double
	
	Public circleA,	circleB, circleR, circleS As Double
	Public circleJ As Int
End Sub

'Initializes the object. You can add parameters to this method if needed.
Public Sub Initialize
	
End Sub

Public Sub calc(pdata As List)
	'Dim i,iter,IterMAX=99 As Int
	Dim iter,IterMAX=99 As Int
    
	Dim Xi,Yi,Zi As Double
	Dim Mz,Mxy,Mxx,Myy,Mxz,Myz,Mzz,Cov_xy,Var_z As Double
	Dim A0,A1,A2,A22,A3,A33 As Double
	Dim Dy,xnew,x,ynew,y As Double
	Dim DET,Xcenter,Ycenter As Double
	
	dataMeanX = 0
	dataMeanY = 0
	For Each xy() As Double In pdata
		dataMeanX = dataMeanX + xy(0)
		dataMeanY = dataMeanY + xy(1)
	Next
	dataN = pdata.Size
	dataMeanX = dataMeanX / dataN
	dataMeanY = dataMeanY / dataN

	For Each xy() As Double In pdata
		Xi = xy(0) - dataMeanX'     centered x-coordinates
		Yi = xy(1) - dataMeanY'     centered y-coordinates
		Zi = Xi*Xi + Yi*Yi
        
		Mxy = Mxy+ Xi*Yi
		Mxx = Mxx+ Xi*Xi
		Myy = Myy+ Yi*Yi
		Mxz = Mxz+ Xi*Zi
		Myz = Myz+ Yi*Zi
		Mzz = Mzz+ Zi*Zi
	Next
		
	Mxx = Mxx / dataN
	Myy = Myy / dataN
	Mxy = Mxy / dataN
	Mxz = Mxz / dataN
	Myz = Myz / dataN
	Mzz = Mzz / dataN
    
	'      computing coefficients of the characteristic polynomial
    
	Mz = Mxx + Myy
	Cov_xy = Mxx*Myy - Mxy*Mxy
	Var_z = Mzz - Mz*Mz
	A3 = 4*Mz
	A2 = -3*Mz*Mz - Mzz
	A1 = Var_z*Mz + 4*Cov_xy*Mz - Mxz*Mxz - Myz*Myz
	A0 = Mxz*(Mxz*Myy - Myz*Mxy) + Myz*(Myz*Mxx - Mxz*Mxy) - Var_z*Cov_xy
	A22 = A2 + A2
	A33 = A3 + A3 + A3

	'    finding the root of the characteristic polynomial
	'    using Newton's method starting at x=0
	'     (it Is guaranteed To converge To the right root)
    
	x=0
	y=A0
	For iter = 0 To IterMAX-1 'For (x=0.,y=A0,iter=0  iter<IterMAX  iter++)  ' usually, 4-6 iterations are enough
	
		Dy = A1 + x*(A22 + A33*x)
		xnew = x - y/Dy
		If ((xnew == x) Or IsNaN(xnew) Or IsInFinity(xnew)) Then Exit 'if ((xnew == x) || (!isfinite(xnew))) break;
		ynew = A0 + xnew*(A1 + xnew*(A2 + xnew*A3))
		If (Abs(ynew)>=Abs(y))  Then Exit
		x = xnew
		y = ynew
	Next
     
	'       computing paramters of the fitting circle
    
	DET = x*x - x*Mz + Cov_xy
	Xcenter = (Mxz*(Myy - x) - Myz*Mxy)/DET/2
	Ycenter = (Myz*(Mxx - x) - Mxz*Mxy)/DET/2

	'       assembling the output

	circleA = Xcenter + dataMeanX
	circleB = Ycenter + dataMeanY
	circleR = Sqrt(Xcenter*Xcenter + Ycenter*Ycenter + Mz)
	'circleS = Sigma(data,circle)
	circleJ = iter   '  Return the number of iterations, too
    
End Sub

Private Sub IsNaN(z As Double) As Boolean
	Return z <> z
End Sub

Private Sub IsInFinity(z As Double) As Boolean
	Return z =	(1/0)
End Sub